import typing as t
from attrs import define, field

from skoll.result import is_ok
from skoll.domain.base import Object
from skoll.domain.primitives import ID, DateTime, Locale, Map, Timezone


__all__ = [
    "Message",
    "MsgClient",
    "RawMessage",
    "MsgContext",
    "MsgPayload",
    "RawMsgClient",
    "MsgClientDevice",
    "RawMsgClientDevice",
]


class RawMsgClientDevice(t.TypedDict):
    duid: t.NotRequired[str]
    model: t.NotRequired[str]
    brand: t.NotRequired[str]
    os_name: t.NotRequired[str]
    os_version: t.NotRequired[str]


class RawMsgClient(t.TypedDict):
    ip_address: t.NotRequired[str]
    locale: t.NotRequired[str]
    timezone: t.NotRequired[str]
    device: t.NotRequired[RawMsgClientDevice]


class RawMessage(t.TypedDict):
    name: str
    source: str
    id: t.NotRequired[str]
    created_at: t.NotRequired[int]
    context: t.NotRequired[RawMsgClient]
    payload: t.NotRequired[dict[str, t.Any]]


@define(frozen=True, kw_only=True, slots=True)
class MsgClientDevice(Object):

    duid: ID | None = None
    model: str | None = None
    brand: str | None = None
    os_name: str | None = None
    os_version: str | None = None

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True)
class MsgClient(Object):

    ip_address: str | None = None
    locale: Locale = field(factory=Locale.default)
    timezone: Timezone = field(factory=Timezone.default)
    device: MsgClientDevice = field(factory=MsgClientDevice.default)

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True)
class MsgContext(Object):

    user_id: ID | None = None
    span_id: ID = field(factory=ID.new)
    trace_id: ID = field(factory=ID.new)
    extra: Map = field(factory=Map.empty)
    client: MsgClient = field(factory=MsgClient.default)

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True)
class MsgPayload(Object):

    @classmethod
    def default(cls) -> t.Self:
        return cls()


@define(frozen=True, kw_only=True, slots=True, eq=False)
class Message(Object):

    name: str
    source: str
    id: ID = field(factory=ID.new)
    created_at: DateTime = field(factory=DateTime.now)
    payload: MsgPayload = field(factory=MsgPayload.default)
    context: MsgContext = field(factory=MsgContext.default)

    @t.override
    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return other.__hash__() == self.__hash__()

    @t.override
    def __ne__(self, other: t.Any) -> bool:
        return not self == other

    @t.override
    def __hash__(self) -> int:
        return hash(self.id.serialize())

    @classmethod
    def from_raw(cls, raw: RawMessage) -> t.Self | None:
        res = cls.create(raw)
        if is_ok(res):
            return res.value
        return None



import json
import typing as t
import collections.abc as c

from nats.aio.msg import Msg
from attrs import define, field
from skoll.utils import default_ssl
from nats.js import JetStreamContext
from skoll.errors import InternalError
from skoll.domain import Message, ID, RawMessage
from skoll.result import Result, fail, ok, is_ok
from nats.aio.client import Client as NatsClient
from skoll.application import Mediator, Subscriber, Service
from nats.aio.subscription import Subscription as NSubscription

from .utils import run_callback

__all__ = ["NatsMediator"]


@define(kw_only=True, slots=True)
class NatsMediator(Mediator):

    creds: str
    servers: list[str]
    _js: JetStreamContext | None = None
    nc: NatsClient = field(init=False, factory=lambda: NatsClient())
    subscriptions: dict[str, list[NSubscription]] = field(factory=dict)

    @property
    def js(self) -> JetStreamContext:
        if self._js is None:
            raise InternalError(debug={"message": "JetStream context is not initialized"})
        return self._js

    @t.override
    async def subscribe(self, service: Service) -> ID:
        if not self.nc.is_connected:
            raise InternalError(debug={"message": "Attempt to subscribe before nats client is connected"})
        sub_id = ID.new()
        subscribtions: list[NSubscription] = []
        for subscriber in service.subscribers:
            if subscriber.js_stream is not None:
                sub = await self.js.subscribe(
                    manual_ack=True,
                    subject=subscriber.topic,
                    stream=subscriber.js_stream,
                    cb=wrap_callback(subscriber),
                    durable=subscriber.service_name,
                    queue=subscriber.service_name if subscriber.queued else None,
                )
            else:
                sub = await self.nc.subscribe(
                    subject=subscriber.topic,
                    cb=wrap_callback(subscriber),
                    queue=subscriber.service_name if subscriber.queued else "",
                )
            subscribtions.append(sub)
        self.subscriptions[sub_id.value] = subscribtions
        return sub_id

    @t.override
    async def unsubscribe(self, id: ID) -> None:
        try:
            subscriptions = self.subscriptions.get(id.value, [])
            for sub in subscriptions:
                await sub.unsubscribe(limit=0)
            del self.subscriptions[id.value]
        except Exception as e:
            raise InternalError.from_exception(e)

    @t.override
    async def connect(self) -> bool:
        try:
            if self.nc.is_connected or self.nc.is_reconnecting:
                return True
            await self.nc.connect(
                tls=default_ssl, servers=self.servers, max_reconnect_attempts=-1, user_credentials=self.creds
            )
            self._js = self.nc.jetstream()
            return True
        except Exception as e:
            print(InternalError.from_exception(e).serialize())
            return False

    @t.override
    async def disconnect(self) -> None:
        if self.nc.is_connected:
            await self.nc.drain()

    @t.override
    async def publish(self, msg: Message | RawMessage) -> None:
        try:
            subject, payload = (msg.name, msg.serialize()) if isinstance(msg, Message) else (msg["name"], msg)
            await self.nc.publish(subject, json.dumps(payload).encode("utf-8"))
        except Exception as e:
            print(InternalError.from_exception(e).serialize())

    @t.override
    async def request(self, msg: Message | RawMessage) -> Result[t.Any]:
        try:
            subject, payload = (msg.name, msg.serialize()) if isinstance(msg, Message) else (msg["name"], msg)
            res = await self.nc.request(subject, json.dumps(payload).encode("utf-8"), timeout=5)
            return ok(json.loads(res.data.decode("utf-8")))
        except TimeoutError as e:
            return fail(InternalError.from_exception(e, extra={"message": f"Request timed out"}))
        except Exception as e:
            print(InternalError.from_exception(e).serialize())
            return fail(InternalError.from_exception(e))


def wrap_callback(subscriber: Subscriber[t.Any]) -> t.Callable[[Msg], c.Awaitable[None]]:
    async def callback(msg: Msg):
        try:
            raw_msg: RawMessage = json.loads(msg.data.decode("utf-8"))
            result = await run_callback(subscriber, raw_msg)
            if subscriber.will_reply:
                await msg.respond(json.dumps(result.value if is_ok(result) else result.err.serialize()).encode("utf-8"))
            elif subscriber.js_stream is not None:
                if is_ok(result):
                    await msg.ack()
                else:
                    await msg.nak(delay=5)
        except Exception as e:
            print(InternalError.from_exception(e))

    return callback



import os
import jwt
import typing as t

from calendar import timegm
from attrs import define, field
from skoll.errors import InternalError
from skoll.result import Result, fail, ok
from datetime import datetime as dt, timedelta, UTC


__all__ = ["DecodedJwtToken", "JwtConfig", "JwtToken"]


class DecodedJwtToken(t.NamedTuple):

    expired: bool = False
    invalid: bool = False
    sub: str | None = None
    kind: str | None = None
    extra: dict[str, t.Any] | None = None


@define(frozen=True, slots=True, kw_only=True)
class JwtConfig:

    issuer: str = field(factory=lambda: os.getenv("JWT_TOKEN_ISSUER", ""))
    encode_key: str = field(factory=lambda: os.getenv("JWT_HASH_SECRET", ""))
    decode_key: str = field(factory=lambda: os.getenv("JWT_HASH_SECRET", ""))
    audience: str = field(factory=lambda: os.getenv("JWT_TOKEN_AUDIENCE", ""))
    algorithm: str = field(factory=lambda: os.getenv("JWT_ALGORITHM", "HS256"))


@define(frozen=True, slots=True, kw_only=True)
class JwtToken:

    config: JwtConfig = field(factory=lambda: JwtConfig())

    def new(self, sub: str, kind: str, duration_min: int, extra: dict[str, t.Any] | None = None) -> Result[str]:
        try:
            iat = dt.now(tz=UTC)
            exp = iat + timedelta(minutes=duration_min)
            payload = {
                "sub": sub,
                "kind": kind,
                "extra": extra or {},
                "iss": self.config.issuer,
                "aud": self.config.audience,
                "exp": timegm(exp.utctimetuple()),
                "iat": timegm(iat.utctimetuple()),
            }
            return ok(value=jwt.encode(payload=payload, key=self.config.encode_key, algorithm=self.config.algorithm))
        except Exception as e:
            return fail(err=InternalError.from_exception(e))

    def decode(self, token: str, kind: str) -> DecodedJwtToken:
        try:
            params: dict[str, t.Any] = {
                "jwt": token,
                "issuer": self.config.issuer,
                "key": self.config.decode_key,
                "audience": self.config.audience,
                "algorithms": [self.config.algorithm],
            }
            payload = jwt.decode(**params)
            kind, sub, extra = payload.get("kind", ""), payload.get("sub", ""), payload.get("extra", {})
            if kind != kind or not sub:
                return DecodedJwtToken(invalid=True)
            return DecodedJwtToken(sub=sub, kind=kind, extra=extra)
        except jwt.ExpiredSignatureError:
            return DecodedJwtToken(expired=True)
        except Exception:
            return DecodedJwtToken(invalid=True)



import typing as t

from skoll.application import Subscriber
from skoll.domain import Message, RawMessage
from skoll.result import Result, fail, is_fail
from skoll.utils import call_with_dependencies
from skoll.errors import Error, InternalError, ValidationFailed

__all__ = ["run_callback"]


async def run_callback(subscriber: Subscriber[t.Any], message: Message | RawMessage) -> Result[t.Any]:
    topic = message.name if isinstance(message, Message) else message.get("name")
    try:
        msg = get_message(subscriber.msg_type, message)
        return await call_with_dependencies(subscriber.callback, {subscriber.msg_arg: msg})
    except Error as err:
        return fail(err=err)
    except Exception as exc:
        return fail(err=InternalError.from_exception(exc, extra={"subject": topic, "message": message}))


def get_message(cls: type[Message], message: Message | RawMessage) -> Message:
    if isinstance(message, Message):
        return message
    res = cls.create(raw=message)
    if is_fail(res):
        raise ValidationFailed(errors=res.err.errors)

    return res.value



import typing as t
from json import load
from skoll.domain import Locale


class I18N:

    base_path: str
    translations: dict[str, dict[str, str]]

    def __init__(self, base_path: str):

        self.translations = {}
        self.base_path = base_path

    def translate(self, key: str, locale: Locale, vars: dict[str, t.Any] | None = None) -> str:
        vars = vars or {}
        name = f"{locale.value.lower().replace('-', '_')}.json"

        if locale.value not in self.translations:
            with open(f"{self.base_path}/{name}", "r") as f:
                self.translations[locale.value] = load(f)

        text = self.translations[locale.value].get(key, f"Translation of <<{key}>> not found for <<{locale.value}>>")
        for k, v in vars.items():
            text = text.replace(f"<<{k}>>", str(v))
        return text



import typing as t
import collections.abc as c

from attrs import define, field
from skoll.result import Result
from abc import ABC, abstractmethod
from skoll.domain import ID, DateTime, Message
from inspect import iscoroutinefunction, signature


__all__ = [
    "Service",
    "Criteria",
    "ListPage",
    "Subscriber",
    "ListCriteria",
    "AuthzWriteChange",
    "AuthzPrecondition",
    "AuthzLookupResult",
    "SubscriberCallback",
    "AuthzWriteOperation",
    "AuthzPreconditionOperation",
]


class ListPage[T](t.NamedTuple):

    items: list[T]
    cursor: str | None = None


class SQLCriteria(t.NamedTuple):
    query: str
    params: list[t.Any]


@define(frozen=True, kw_only=True, slots=True)
class Criteria(ABC):

    uid: ID | None = None

    @property
    @abstractmethod
    def as_sql(self) -> SQLCriteria:
        raise NotImplementedError("Subclasses must implement this method")


@define(frozen=True, kw_only=True, slots=True)
class ListCriteria(Criteria, ABC):

    cursor: str | None = None
    limit: int = field(default=100)


type AuthzWriteOperation = t.Literal["SET", "DELETE"]
type AuthzPrecondition = tuple[AuthzPreconditionOperation, str]
type AuthzPreconditionOperation = t.Literal["MUST_MATCH", "MUST_NOT_MATCH"]
type AuthzWriteChange = tuple[AuthzWriteOperation, str, tuple[str, dict[str, t.Any]] | None, DateTime | None]
type SubscriberCallback[T: Message] = t.Callable[t.Concatenate[T, ...], c.Coroutine[t.Any, t.Any, Result[t.Any]]]


class AuthzLookupResult(t.NamedTuple):
    uids: list[str]
    cursor: str | None = None


@define(frozen=True, kw_only=True, slots=True)
class Subscriber[T: Message]:

    topic: str
    msg_arg: str
    queued: bool
    will_reply: bool
    service_name: str
    msg_type: type[Message]
    js_stream: str | None = None
    callback: SubscriberCallback[T]


@define(kw_only=True, slots=True, frozen=True)
class Service:

    name: str
    subscribers: list[Subscriber[t.Any]] = field(factory=list)

    def _add[T: Message](
        self, topic: str, will_reply: bool, queued: bool, callback: SubscriberCallback[T], js_stream: str | None = None
    ):
        first_arg = list(signature(callback).parameters.values())[0]
        if not issubclass(first_arg.annotation, Message) or not iscoroutinefunction(callback):
            raise TypeError(
                f"Service subscriber @on/@reply must be a coroutine and with first argument being a subclass of Message"
            )
        self.subscribers.append(
            Subscriber(
                topic=topic,
                queued=queued,
                callback=callback,
                js_stream=js_stream,
                will_reply=will_reply,
                msg_arg=first_arg.name,
                service_name=self.name,
                msg_type=first_arg.annotation,
            )
        )

    def on(self, topic: str, queued: bool = False, stream: str | None = None):
        def decorator[T: Message](callback: SubscriberCallback[T]):
            self._add(topic, will_reply=False, queued=queued, callback=callback, js_stream=stream)
            return callback

        return decorator

    def reply(self, topic: str):
        def decorator[T: Message](callback: SubscriberCallback[T]):
            self._add(topic, will_reply=True, queued=True, callback=callback)
            return callback

        return decorator
